# SPDX-FileCopyrightText: Copyright (c) 2025, NVIDIA CORPORATION. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import base64
import copy
import datetime
from dataclasses import dataclass
from typing import Optional

import yaml
from omegaconf import DictConfig, OmegaConf

from nemo_evaluator_launcher.common.logging_utils import logger


@dataclass(frozen=True)
class CmdAndReadableComment:
    """See the comment to `_yaml_to_echo_command`."""

    # Actual command. Might include hard-to-debug elements such as base64-encoded
    # configs.
    cmd: str
    # A debuggale readable comment that can be passed along for accompanying
    # the actual command
    debug: str


def _yaml_to_echo_command(
    yaml_str: str, filename: str = "config_ef.yaml"
) -> CmdAndReadableComment:
    """Create a safe (see below) echo command saving a yaml to file.

    Safety in this context means the ability to pass such echo command through the
    `bash -c '...'` boundaries for example.

    Naturally, enconding with base64 creates debuggability issues. For that, the second
    output of the function is the yaml string with bash comment signs prepended.
    """
    yaml_str_b64 = base64.b64encode(yaml_str.encode("utf-8")).decode("utf-8")
    debug_str = "\n".join(
        [f"# Contents of {filename}"] + ["# " + s for s in yaml_str.splitlines()]
    )
    return CmdAndReadableComment(
        cmd=f'echo "{yaml_str_b64}" | base64 -d > {filename}', debug=debug_str
    )


def _set_nested_optionally_overriding(
    d: dict, keys: list[str], val: object, *, override_if_exists: bool = False
):
    """Sets d[...keys....] = value, creating keys all the way"""
    temp = d
    for key in keys[:-1]:
        temp = temp.setdefault(key, {})
    if override_if_exists or keys[-1] not in temp:
        temp[keys[-1]] = val


def get_eval_factory_config(
    cfg: DictConfig,
    user_task_config: DictConfig,
) -> dict:
    """Extract config fields for eval factory.

    This function extracts the config field similar to how overrides are handled.

    Overrides will be start to be deprecated (or not, but at least a warning will be logged).
    """

    if cfg.evaluation.get("overrides") or user_task_config.get("overrides"):
        # TODO(agronskiy): start removing overrides, test `test_start_deprecating_overrides`
        # will start failing soon.
        logger.warning(
            "We are deprecating using old-style dot-delimited overrides "
            "in favour of `nemo_evaluator_config` field. Please check "
            "the documentation."
        )

    logger.debug("Getting nemo evaluator merged config")
    # Extract config fields similar to overrides - convert to basic Python types first
    # Support both new and old format for backward compatibility
    cfg_config = cfg.evaluation.get("nemo_evaluator_config") or cfg.evaluation.get(
        "config", {}
    )
    user_config = user_task_config.get("nemo_evaluator_config") or user_task_config.get(
        "config", {}
    )

    # Convert OmegaConf objects to basic Python types
    if cfg_config:
        cfg_config = OmegaConf.to_container(cfg_config, resolve=True)
    if user_config:
        user_config = OmegaConf.to_container(user_config, resolve=True)

    # Merge the configs
    merged_nemo_evaluator_config: dict = OmegaConf.to_container(
        OmegaConf.merge(cfg_config, user_config)
    )

    logger.debug(
        "Merged nemo evaluator config, not final",
        source_global_cfg=cfg_config,
        source_task_config=user_config,
        result=merged_nemo_evaluator_config,
    )

    return merged_nemo_evaluator_config


def get_eval_factory_command(
    cfg: DictConfig, user_task_config: DictConfig, task_definition: dict
) -> CmdAndReadableComment:
    merged_nemo_evaluator_config = get_eval_factory_config(
        cfg,
        user_task_config,
    )

    # We now prepare the config to be passed to `nemo-evaluator` command.
    _set_nested_optionally_overriding(
        merged_nemo_evaluator_config,
        ["target", "api_endpoint", "url"],
        get_endpoint_url(
            cfg,
            merged_nemo_evaluator_config=merged_nemo_evaluator_config,
            endpoint_type=task_definition["endpoint_type"],
        ),
    )
    _set_nested_optionally_overriding(
        merged_nemo_evaluator_config,
        ["target", "api_endpoint", "model_id"],
        get_served_model_name(cfg),
    )
    _set_nested_optionally_overriding(
        merged_nemo_evaluator_config,
        ["target", "api_endpoint", "type"],
        task_definition["endpoint_type"],
    )
    _set_nested_optionally_overriding(
        merged_nemo_evaluator_config,
        ["config", "type"],
        task_definition["task"],
    )
    _set_nested_optionally_overriding(
        merged_nemo_evaluator_config,
        ["config", "output_dir"],
        "/results",
    )
    _set_nested_optionally_overriding(
        merged_nemo_evaluator_config,
        ["target", "api_endpoint", "api_key"],
        "API_KEY",
    )

    create_file_cmd = _yaml_to_echo_command(
        yaml.safe_dump(merged_nemo_evaluator_config), "config_ef.yaml"
    )
    eval_command = (
        "cmd=$(command -v nemo-evaluator >/dev/null 2>&1 && echo nemo-evaluator || echo eval-factory) "
        + "&& $cmd run_eval --run_config config_ef.yaml"
    )

    # NOTE: see note and test about deprecating that.
    overrides = copy.deepcopy(dict(cfg.evaluation.get("overrides", {})))
    overrides.update(dict(user_task_config.get("overrides", {})))
    # NOTE(dfridman): Temporary fix to make sure that the overrides arg is not split into multiple lines.
    # Consider passing a JSON object on Eval Factory side
    overrides = {
        k: (v.strip("\n") if isinstance(v, str) else v) for k, v in overrides.items()
    }
    overrides_str = ",".join([f"{k}={v}" for k, v in overrides.items()])
    if overrides_str:
        eval_command = f"{eval_command} --overrides {overrides_str}"

    # We return both the command and the debugging base64-decoded strings, useful
    # for exposing when building scripts.
    return CmdAndReadableComment(
        cmd=create_file_cmd.cmd + " && " + eval_command, debug=create_file_cmd.debug
    )


def get_endpoint_url(
    cfg: DictConfig,
    merged_nemo_evaluator_config: dict,
    endpoint_type: str,
) -> str:
    def apply_url_override(url: str) -> str:
        """Apply user URL override if provided."""
        nemo_evaluator_config_url = (
            merged_nemo_evaluator_config.get("target", {})
            .get("api_endpoint", {})
            .get("url", None)
        )

        if nemo_evaluator_config_url:
            return nemo_evaluator_config_url

        # Being deprecated, see `get_eval_factory_config` message.
        overrides_old_style_url = merged_nemo_evaluator_config.get("overrides", {}).get(
            "target.api_endpoint.url", None
        )
        if overrides_old_style_url:
            return overrides_old_style_url

        return url

    if cfg.deployment.type == "none":
        # For deployment: none, use target URL regardless of executor type
        if OmegaConf.is_missing(cfg.target.api_endpoint, "url"):
            raise ValueError(
                "API endpoint URL is not set. Add `target.api_endpoint.url` to your config "
                "OR override via CLI"
            )
        return apply_url_override(cfg.target.api_endpoint.url)

    elif (
        hasattr(cfg, "target")
        and hasattr(cfg.target, "api_endpoint")
        and hasattr(cfg.target.api_endpoint, "url")
        and not OmegaConf.is_missing(cfg.target.api_endpoint, "url")
    ):
        # For Lepton executor with dynamically set target URL
        return apply_url_override(cfg.target.api_endpoint.url)

    else:
        # Local executor - use localhost
        endpoint_uri = cfg.deployment.endpoints[endpoint_type]
        endpoint_url = f"http://127.0.0.1:{cfg.deployment.port}{endpoint_uri}"
        return endpoint_url


def get_health_url(cfg: DictConfig, endpoint_url: str) -> str:
    if cfg.deployment.type == "none":
        logger.warning("Using endpoint URL as health URL", will_be_used=endpoint_url)
        return endpoint_url  # TODO(public release) is using model url as health url OK?
    health_uri = cfg.deployment.endpoints["health"]
    health_url = f"http://127.0.0.1:{cfg.deployment.port}{health_uri}"
    return health_url


def get_served_model_name(cfg: DictConfig) -> str:
    if cfg.deployment.type == "none":
        return str(cfg.target.api_endpoint.model_id)
    else:
        return str(cfg.deployment.served_model_name)


def get_api_key_name(cfg: DictConfig) -> str | None:
    res = cfg.get("target", {}).get("api_endpoint", {}).get("api_key_name", None)
    return str(res) if res else None


def get_timestamp_string(include_microseconds: bool = True) -> str:
    """Get timestamp in format YYYYmmdd_HHMMSS_ffffff."""
    dt = datetime.datetime.now()
    fmt = "%Y%m%d_%H%M%S"
    if include_microseconds:
        fmt += "_%f"
    dts = datetime.datetime.strftime(dt, fmt)
    return dts


def get_eval_factory_dataset_size_from_run_config(run_config: dict) -> Optional[int]:
    config = run_config["config"]
    limit_samples = config["params"].get("limit_samples", None)
    if limit_samples is not None:
        return int(limit_samples)

    # TODO(dfridman): Move `dataset_size` values to the corresponding `framework.yaml` in Eval Factory.
    dataset_sizes = {
        ("lm-evaluation-harness", "ifeval"): 541,
        ("simple_evals", "gpqa_diamond"): 198,
        ("simple_evals", "gpqa_diamond_nemo"): 198,
        ("simple_evals", "AA_math_test_500"): 500,
        ("simple_evals", "math_test_500_nemo"): 500,
        ("simple_evals", "aime_2024_nemo"): 30,
        ("simple_evals", "AA_AIME_2024"): 30,
        ("simple_evals", "aime_2025_nemo"): 30,
        ("simple_evals", "AIME_2025"): 30,
        ("simple_evals", "humaneval"): 164,
        ("simple_evals", "mmlu"): 14042,
        ("simple_evals", "mmlu_pro"): 12032,
        ("bigcode-evaluation-harness", "mbpp"): 500,
        ("bigcode-evaluation-harness", "humaneval"): 164,
        ("livecodebench", "livecodebench_0724_0125"): 315,
        ("livecodebench", "livecodebench_0824_0225"): 279,
        ("hle", "hle"): 2684,
        ("scicode", "aa_scicode"): 338,
    }
    dataset_size = dataset_sizes.get((run_config["framework_name"], config["type"]))
    if dataset_size is None:
        return None
    else:
        dataset_size = int(dataset_size)

    downsampling_ratio = (
        config["params"].get("extra", {}).get("downsampling_ratio", None)
    )
    if downsampling_ratio is not None:
        dataset_size = int(round(dataset_size * downsampling_ratio))

    n_samples = int(config["params"].get("extra", {}).get("n_samples", 1))
    dataset_size *= n_samples
    return dataset_size



# SPDX-FileCopyrightText: Copyright (c) 2025, NVIDIA CORPORATION. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Version command for nemo-evaluator-launcher."""

import importlib
from dataclasses import dataclass

from nemo_evaluator_launcher import __package_name__, __version__


@dataclass
class Cmd:
    """Show version information for nemo-evaluator-launcher and internal packages."""

    def execute(self) -> None:
        """Execute the version command."""
        print(f"{__package_name__}: {__version__}")

        # Check for internal package
        try:
            internal_module = importlib.import_module(
                "nemo_evaluator_launcher_internal"
            )
            # Try to get version from internal package
            try:
                internal_version = getattr(internal_module, "__version__", None)
                if internal_version:
                    print(f"nemo-evaluator-launcher-internal: {internal_version}")
                else:
                    print(
                        "nemo-evaluator-launcher-internal: available (version unknown)"
                    )
            except Exception:
                print("nemo-evaluator-launcher-internal: available (version unknown)")
        except ImportError:
            # Internal package not available - this is expected in many cases
            pass
        except Exception as e:
            print(f"nemo-evaluator-launcher-internal: error loading ({e})")



# SPDX-FileCopyrightText: Copyright (c) 2025, NVIDIA CORPORATION. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Shared utilities for metrics and configuration handling."""

import json
import re
import subprocess
from pathlib import Path
from typing import Any, Callable, Dict, List, Tuple

import yaml

from nemo_evaluator_launcher.common.execdb import JobData
from nemo_evaluator_launcher.common.logging_utils import logger
from nemo_evaluator_launcher.common.mapping import (
    get_task_from_mapping,
    load_tasks_mapping,
)

# =============================================================================
# ARTIFACTS
# =============================================================================

# Artifacts to be logged by default
REQUIRED_ARTIFACTS = ["results.yml", "eval_factory_metrics.json"]
OPTIONAL_ARTIFACTS = ["omni-info.json"]


def get_relevant_artifacts() -> List[str]:
    """Get relevant artifacts (required + optional)."""
    return REQUIRED_ARTIFACTS + OPTIONAL_ARTIFACTS


def validate_artifacts(artifacts_dir: Path) -> Dict[str, Any]:
    """Check which artifacts are available."""
    if not artifacts_dir or not artifacts_dir.exists():
        return {
            "can_export": False,
            "missing_required": REQUIRED_ARTIFACTS.copy(),
            "missing_optional": OPTIONAL_ARTIFACTS.copy(),
            "message": "Artifacts directory not found",
        }

    missing_required = [
        f for f in REQUIRED_ARTIFACTS if not (artifacts_dir / f).exists()
    ]
    missing_optional = [
        f for f in OPTIONAL_ARTIFACTS if not (artifacts_dir / f).exists()
    ]
    can_export = len(missing_required) == 0

    message_parts = []
    if missing_required:
        message_parts.append(f"Missing required: {', '.join(missing_required)}")
    if missing_optional:
        message_parts.append(f"Missing optional: {', '.join(missing_optional)}")

    return {
        "can_export": can_export,
        "missing_required": missing_required,
        "missing_optional": missing_optional,
        "message": (
            ". ".join(message_parts) if message_parts else "All artifacts available"
        ),
    }


def get_available_artifacts(artifacts_dir: Path) -> List[str]:
    """Get list of artifacts available in artifacts directory."""
    if not artifacts_dir or not artifacts_dir.exists():
        return []
    return [
        filename
        for filename in get_relevant_artifacts()
        if (artifacts_dir / filename).exists()
    ]


# =============================================================================
# METRICS EXTRACTION
# =============================================================================


class MetricConflictError(Exception):
    """Raised when attempting to set the same metric key with a different value."""


def extract_accuracy_metrics(
    job_data: JobData, get_job_paths_func: Callable, log_metrics: List[str] = None
) -> Dict[str, float]:
    """Extract accuracy metrics from job results."""
    try:
        paths = get_job_paths_func(job_data)
        artifacts_dir = _get_artifacts_dir(paths)

        if not artifacts_dir or not artifacts_dir.exists():
            logger.warning(f"Artifacts directory not found for job {job_data.job_id}")
            return {}

        # Prefer results.yml, but also merge JSON metrics to avoid missing values
        metrics: Dict[str, float] = {}
        results_yml = artifacts_dir / "results.yml"
        if results_yml.exists():
            yml_metrics = _extract_from_results_yml(results_yml)
            if yml_metrics:
                metrics.update(yml_metrics)

        # Merge in JSON metrics (handles tasks that only emit JSON or extra fields)
        json_metrics = _extract_from_json_files(artifacts_dir)
        for k, v in json_metrics.items():
            metrics.setdefault(k, v)

        # Filter metrics if specified
        if log_metrics:
            filtered_metrics = {}
            for metric_name, metric_value in metrics.items():
                if any(filter_key in metric_name.lower() for filter_key in log_metrics):
                    filtered_metrics[metric_name] = metric_value
            return filtered_metrics

        return metrics

    except Exception as e:
        logger.error(f"Failed to extract metrics for job {job_data.job_id}: {e}")
        return {}


# =============================================================================
# CONFIG EXTRACTION
# =============================================================================


def extract_exporter_config(
    job_data: JobData, exporter_name: str, constructor_config: Dict[str, Any] = None
) -> Dict[str, Any]:
    """Extract and merge exporter configuration from multiple sources."""
    config = {}

    # root-level `export.<exporter-name>`
    if job_data.config:
        export_block = (job_data.config or {}).get("export", {})
        yaml_config = (export_block or {}).get(exporter_name, {})
        if yaml_config:
            config.update(yaml_config)

    # From webhook metadata (if triggered by webhook)
    if "webhook_metadata" in job_data.data:
        webhook_data = job_data.data["webhook_metadata"]
        webhook_config = {
            "triggered_by_webhook": True,
            "webhook_source": webhook_data.get("webhook_source", "unknown"),
            "source_artifact": f"{webhook_data.get('artifact_name', 'unknown')}:{webhook_data.get('artifact_version', 'unknown')}",
            "config_source": webhook_data.get("config_file", "unknown"),
        }
        if exporter_name == "wandb" and webhook_data.get("webhook_source") == "wandb":
            wandb_specific = {
                "entity": webhook_data.get("entity"),
                "project": webhook_data.get("project"),
                "run_id": webhook_data.get("run_id"),
            }
            webhook_config.update({k: v for k, v in wandb_specific.items() if v})
        config.update(webhook_config)

    # allows CLI overrides
    if constructor_config:
        config.update(constructor_config)

    return config


# =============================================================================
# JOB DATA EXTRACTION
# =============================================================================


def get_task_name(job_data: JobData) -> str:
    """Get task name from job data."""
    if "." in job_data.job_id:
        try:
            idx = int(job_data.job_id.split(".")[-1])
            return job_data.config["evaluation"]["tasks"][idx]["name"]
        except Exception:
            return f"job_{job_data.job_id}"
    return "all_tasks"


def get_model_name(job_data: JobData, config: Dict[str, Any] = None) -> str:
    """Extract model name from config or job data."""
    if config and "model_name" in config:
        return config["model_name"]

    job_config = job_data.config or {}
    model_sources = [
        job_config.get("target", {}).get("api_endpoint", {}).get("model_id"),
        job_config.get("deployment", {}).get("served_model_name"),
        job_data.data.get("served_model_name"),
        job_data.data.get("model_name"),
        job_data.data.get("model_id"),
    ]

    for source in model_sources:
        if source:
            return str(source)

    return f"unknown_model_{job_data.job_id}"


def get_pipeline_id(job_data: JobData) -> str:
    """Get pipeline ID for GitLab jobs."""
    return job_data.data.get("pipeline_id") if job_data.executor == "gitlab" else None


def get_benchmark_info(job_data: JobData) -> Dict[str, str]:
    """Get harness and benchmark info from mapping."""
    try:
        task_name = get_task_name(job_data)
        if task_name in ["all_tasks", f"job_{job_data.job_id}"]:
            return {"harness": "unknown", "benchmark": task_name}

        # Use mapping to get harness info
        mapping = load_tasks_mapping()
        task_definition = get_task_from_mapping(task_name, mapping)
        harness = task_definition.get("harness", "unknown")

        # Extract benchmark name (remove harness prefix)
        if "." in task_name:
            benchmark = ".".join(task_name.split(".")[1:])
        else:
            benchmark = task_name

        return {"harness": harness, "benchmark": benchmark}

    except Exception as e:
        logger.warning(f"Failed to get benchmark info: {e}")
        return {"harness": "unknown", "benchmark": get_task_name(job_data)}


def get_container_from_mapping(job_data: JobData) -> str:
    """Get container from mapping."""
    try:
        task_name = get_task_name(job_data)
        if task_name in ["all_tasks", f"job_{job_data.job_id}"]:
            return None

        mapping = load_tasks_mapping()
        task_definition = get_task_from_mapping(task_name, mapping)
        return task_definition.get("container")

    except Exception as e:
        logger.warning(f"Failed to get container from mapping: {e}")
        return None


def get_artifact_root(job_data: JobData) -> str:
    """Get artifact root from job data."""
    bench = get_benchmark_info(job_data)
    h = bench.get("harness", "unknown")
    b = bench.get("benchmark", get_task_name(job_data))
    return f"{h}.{b}"


# =============================================================================
# GITLAB DOWNLOAD
# =============================================================================


def download_gitlab_artifacts(
    paths: Dict[str, Any], export_dir: Path, extract_specific: bool = False
) -> Dict[str, Path]:
    """Download artifacts from GitLab API.

    Args:
        paths: Dictionary containing pipeline_id and project_id
        export_dir: Local directory to save artifacts
        extract_specific: If True, extract individual files; if False, keep as ZIP files

    Returns:
        Dictionary mapping artifact names to local file paths
    """
    raise NotImplementedError("Downloading from gitlab is not implemented")


# =============================================================================
# SSH UTILS
# =============================================================================


# SSH connections directory
CONNECTIONS_DIR = Path.home() / ".nemo-evaluator" / "connections"


def ssh_setup_masters(jobs: Dict[str, JobData]) -> Dict[Tuple[str, str], str]:
    """Start SSH master connections for remote jobs, returns control_paths."""
    remote_pairs: set[tuple[str, str]] = set()
    for jd in jobs.values():
        try:
            # Preferred: explicit 'paths' from job data
            p = (jd.data or {}).get("paths") or {}
            if (
                p.get("storage_type") == "remote_ssh"
                and p.get("username")
                and p.get("hostname")
            ):
                remote_pairs.add((p["username"], p["hostname"]))
                continue
            # Fallback: common slurm fields (works with BaseExporter.get_job_paths)
            d = jd.data or {}
            if jd.executor == "slurm" and d.get("username") and d.get("hostname"):
                remote_pairs.add((d["username"], d["hostname"]))
        except Exception:
            pass

    if not remote_pairs:
        return {}

    CONNECTIONS_DIR.mkdir(parents=True, exist_ok=True)
    control_paths: Dict[Tuple[str, str], str] = {}
    for username, hostname in remote_pairs:
        socket_path = CONNECTIONS_DIR / f"{username}_{hostname}.sock"
        try:
            cmd = [
                "ssh",
                "-N",
                "-f",
                "-o",
                "ControlMaster=auto",
                "-o",
                "ControlPersist=60",
                "-o",
                f"ControlPath={socket_path}",
                f"{username}@{hostname}",
            ]
            subprocess.run(cmd, check=False, capture_output=True)
            control_paths[(username, hostname)] = str(socket_path)
        except Exception as e:
            logger.warning(f"Failed to start SSH master for {username}@{hostname}: {e}")
    return control_paths


def ssh_cleanup_masters(control_paths: Dict[Tuple[str, str], str]) -> None:
    """Clean up SSH master connections from control_paths."""
    for (username, hostname), socket_path in (control_paths or {}).items():
        try:
            cmd = [
                "ssh",
                "-O",
                "exit",
                "-o",
                f"ControlPath={socket_path}",
                f"{username}@{hostname}",
            ]
            subprocess.run(cmd, check=False, capture_output=True)
        except Exception as e:
            logger.warning(f"Failed to stop SSH master for {username}@{hostname}: {e}")

        # Clean up
        try:
            Path(socket_path).unlink(missing_ok=True)
        except Exception as e:
            logger.warning(f"Failed to clean up file: {e}")


def ssh_download_artifacts(
    paths: Dict[str, Any],
    export_dir: Path,
    config: Dict[str, Any] | None = None,
    control_paths: Dict[Tuple[str, str], str] | None = None,
) -> List[str]:
    """Download artifacts/logs via SSH with optional connection reuse."""
    exported_files: List[str] = []
    copy_logs = bool((config or {}).get("copy_logs", False))
    copy_artifacts = bool((config or {}).get("copy_artifacts", True))
    only_required = bool((config or {}).get("only_required", True))

    control_path = None
    if control_paths:
        control_path = control_paths.get((paths["username"], paths["hostname"]))
    ssh_opts = ["-o", f"ControlPath={control_path}"] if control_path else []

    def scp_file(remote_path: str, local_path: Path) -> bool:
        cmd = (
            ["scp"]
            + ssh_opts
            + [
                f"{paths['username']}@{paths['hostname']}:{remote_path}",
                str(local_path),
            ]
        )
        return subprocess.run(cmd, capture_output=True).returncode == 0

    export_dir.mkdir(parents=True, exist_ok=True)

    # Artifacts
    if copy_artifacts:
        art_dir = export_dir / "artifacts"
        art_dir.mkdir(parents=True, exist_ok=True)

        if only_required:
            for artifact in get_relevant_artifacts():
                remote_file = f"{paths['remote_path']}/artifacts/{artifact}"
                local_file = art_dir / artifact
                local_file.parent.mkdir(parents=True, exist_ok=True)
                if scp_file(remote_file, local_file):
                    exported_files.append(str(local_file))
        else:
            # Copy known files individually to avoid subfolders and satisfy tests
            for artifact in get_available_artifacts(paths.get("artifacts_dir", Path())):
                remote_file = f"{paths['remote_path']}/artifacts/{artifact}"
                local_file = art_dir / artifact
                if scp_file(remote_file, local_file):
                    exported_files.append(str(local_file))

    # Logs (top-level only)
    if copy_logs:
        local_logs = export_dir / "logs"
        remote_logs = f"{paths['remote_path']}/logs"
        cmd = (
            ["scp", "-r"]
            + ssh_opts
            + [
                f"{paths['username']}@{paths['hostname']}:{remote_logs}/.",
                str(local_logs),
            ]
        )
        if subprocess.run(cmd, capture_output=True).returncode == 0:
            for p in local_logs.iterdir():
                if p.is_dir():
                    import shutil

                    shutil.rmtree(p, ignore_errors=True)
            exported_files.extend([str(f) for f in local_logs.glob("*") if f.is_file()])

    return exported_files


# =============================================================================
# PRIVATE HELPER FUNCTIONS
# =============================================================================


def _get_artifacts_dir(paths: Dict[str, Any]) -> Path:
    """Get artifacts directory from paths."""
    storage_type = paths.get("storage_type")

    # For SSH-based remote access, artifacts aren't available locally yet
    if storage_type == "remote_ssh":
        return None

    # For all local access (local_filesystem, remote_local, gitlab_ci_local)
    # return the artifacts_dir from paths
    return paths.get("artifacts_dir")


def _extract_metrics_from_results(results: dict) -> Dict[str, float]:
    """Extract metrics from a 'results' dict (with optional 'groups'/'tasks')."""
    metrics: Dict[str, float] = {}
    for section in ["groups", "tasks"]:
        section_data = results.get(section)
        if isinstance(section_data, dict):
            for task_name, task_data in section_data.items():
                if isinstance(task_data, dict) and "metrics" in task_data:
                    task_metrics = _extract_task_metrics(
                        task_name, task_data["metrics"]
                    )
                    _safe_update_metrics(
                        target=metrics,
                        source=task_metrics,
                        context=f" while extracting results for task '{task_name}'",
                    )
    return metrics


def _extract_from_results_yml(results_yml: Path) -> Dict[str, float]:
    """Extract metrics from results.yml file."""
    try:
        with open(results_yml, "r", encoding="utf-8") as f:
            data = yaml.safe_load(f)
        if not isinstance(data, dict) or "results" not in data:
            return {}
        return _extract_metrics_from_results(data.get("results"))
    except Exception as e:
        logger.warning(f"Failed to parse results.yml: {e}")
        return {}


def _extract_from_json_files(artifacts_dir: Path) -> Dict[str, float]:
    """Extract metrics from individual JSON result files."""
    metrics = {}

    for json_file in artifacts_dir.glob("*.json"):
        if json_file.name in get_relevant_artifacts():
            continue  # Skip known artifact files, focus on task result files

        try:
            with open(json_file, "r", encoding="utf-8") as f:
                data = json.load(f)

            if isinstance(data, dict) and "score" in data:
                task_name = json_file.stem
                metrics[f"{task_name}_score"] = float(data["score"])

        except Exception as e:
            logger.warning(f"Failed to parse {json_file}: {e}")

    return metrics


def _extract_task_metrics(task_name: str, metrics_data: dict) -> Dict[str, float]:
    """Extract metrics from a task's metrics data."""
    extracted = {}
    score_patterns = [
        "acc",
        "accuracy",
        "score",
        "exact_match",
        "f1",
        "em",
        "pass@1",
        "pass@k",
    ]

    for metric_name, metric_data in metrics_data.items():
        # Only extract score-like metrics
        if not any(pattern in metric_name.lower() for pattern in score_patterns):
            continue

        try:
            if isinstance(metric_data, dict):
                if "scores" in metric_data:
                    # Handle nested scores (e.g., mmlu macro/micro)
                    for score_type, score_data in metric_data["scores"].items():
                        if isinstance(score_data, dict) and "value" in score_data:
                            key = f"{task_name}_{metric_name}_{score_type}"
                            _safe_set_metric(
                                container=extracted,
                                key=key,
                                new_value=score_data["value"],
                                context=f" in task '{task_name}'",
                            )
                elif "value" in metric_data:
                    key = f"{task_name}_{metric_name}"
                    _safe_set_metric(
                        container=extracted,
                        key=key,
                        new_value=metric_data["value"],
                        context=f" in task '{task_name}'",
                    )
            elif isinstance(metric_data, (int, float)):
                key = f"{task_name}_{metric_name}"
                _safe_set_metric(
                    container=extracted,
                    key=key,
                    new_value=metric_data,
                    context=f" in task '{task_name}'",
                )
        except (ValueError, TypeError) as e:
            logger.warning(
                f"Failed to extract metric {metric_name} for task {task_name}: {e}"
            )

    return extracted


def _safe_set_metric(
    container: Dict[str, float], key: str, new_value: float, context: str
) -> None:
    """Set a metric into container; raise with details if key exists."""
    if key in container:
        # Allow exact matches; warn and keep existing
        if container[key] == float(new_value):
            logger.warning(
                f"Metric rewrite{context}: '{key}' has identical value; keeping existing. value={container[key]}"
            )
            return
        # Different value is an error we want to surface distinctly
        raise MetricConflictError(
            f"Metric key collision{context}: '{key}' already set. existing={container[key]} new={new_value}"
        )
    container[key] = float(new_value)


def _safe_update_metrics(
    target: Dict[str, float], source: Dict[str, float], context: str
) -> None:
    """Update target from source safely, raising on collisions with detailed values."""
    for k, v in source.items():
        _safe_set_metric(target, k, v, context)


# =============================================================================
# MLFLOW FUNCTIONS
# =============================================================================

# MLflow constants
_MLFLOW_KEY_MAX = 250
_MLFLOW_PARAM_VAL_MAX = 250
_MLFLOW_TAG_VAL_MAX = 5000

_INVALID_KEY_CHARS = re.compile(r"[^/\w.\- ]")
_MULTI_UNDERSCORE = re.compile(r"_+")


def mlflow_sanitize(s: Any, kind: str = "key") -> str:
    """
    Sanitize strings for MLflow logging.

    kind:
      - "key", "metric", "tag_key", "param_key": apply key rules
      - "tag_value": apply tag value rules
      - "param_value": apply param value rules
    """
    s = "" if s is None else str(s)

    if kind in ("key", "metric", "tag_key", "param_key"):
        # common replacements
        s = s.replace("pass@", "pass_at_")
        # drop disallowed chars, collapse underscores, trim
        s = _INVALID_KEY_CHARS.sub("_", s)
        s = _MULTI_UNDERSCORE.sub("_", s).strip()
        return s[:_MLFLOW_KEY_MAX] or "key"

    # values: normalize whitespace, enforce length
    s = s.replace("\n", " ").replace("\r", " ").strip()
    max_len = _MLFLOW_TAG_VAL_MAX if kind == "tag_value" else _MLFLOW_PARAM_VAL_MAX
    return s[:max_len]



# SPDX-FileCopyrightText: Copyright (c) 2025, NVIDIA CORPORATION. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Lepton deployment helper functions for nemo-evaluator-launcher.

Handles Lepton endpoint creation, management, and health checks.
"""

import json
import subprocess
import time
from pathlib import Path
from typing import Any, Dict, Optional

# Import lepton dependencies
from omegaconf import DictConfig

from nemo_evaluator_launcher.common.logging_utils import logger


def deep_merge(base: Dict[Any, Any], override: Dict[Any, Any]) -> Dict[Any, Any]:
    """Deep merge two dictionaries, with override taking precedence."""
    result = base.copy()

    for key, value in override.items():
        if key in result and isinstance(result[key], dict) and isinstance(value, dict):
            result[key] = deep_merge(result[key], value)
        else:
            result[key] = value

    return result


def replace_placeholders(data: Any, replacements: Dict[str, str]) -> Any:
    """Replace placeholders in the data structure."""

    def replace_in_obj(obj: Any) -> Any:
        if isinstance(obj, dict):
            return {k: replace_in_obj(v) for k, v in obj.items()}
        elif isinstance(obj, list):
            return [replace_in_obj(item) for item in obj]
        elif isinstance(obj, str):
            result = obj
            for placeholder, value in replacements.items():
                result = result.replace(f"{{{{{placeholder}}}}}", value)
            return result
        else:
            return obj

    return replace_in_obj(data)


def generate_lepton_spec(cfg: DictConfig) -> Dict[str, Any]:
    """Generate a Lepton endpoint specification from nemo-evaluator-launcher configuration.

    This function creates a layered configuration by merging:
    1. Platform defaults (from execution.lepton_platform.platform_defaults)
    2. Environment settings (from execution.lepton_platform)
    3. Inference engine config (from deployment.* - vllm/sglang settings)
    4. Lepton platform config (from deployment.lepton_config - Lepton-specific settings)

    Args:
        cfg: The nemo-evaluator-launcher configuration object containing all settings.

    Returns:
        Dict containing the Lepton endpoint specification.
    """

    # Step 1: Start with platform defaults from execution config
    platform_defaults = {}
    if hasattr(cfg, "execution") and hasattr(cfg.execution, "lepton_platform"):
        deployment_config = cfg.execution.lepton_platform.get("deployment", {})
        platform_defaults = deployment_config.get("platform_defaults", {})

    base_config = deep_merge({}, platform_defaults)

    # Step 2: Apply deployment-specific settings from execution config
    if hasattr(cfg, "execution") and hasattr(cfg.execution, "lepton_platform"):
        lepton_platform = cfg.execution.lepton_platform
        deployment_config = lepton_platform.get("deployment", {})

        # Add deployment node group as affinity constraint
        deployment_node_group = deployment_config.get("node_group")
        if deployment_node_group:
            if not base_config.get("resource_requirement"):
                base_config["resource_requirement"] = {}
            base_config["resource_requirement"]["affinity"] = {
                "allowed_dedicated_node_groups": [deployment_node_group]
            }

        # Add queue config from platform defaults
        platform_defaults = deployment_config.get("platform_defaults", {})
        if platform_defaults.get("queue_config"):
            base_config["queue_config"] = platform_defaults.get("queue_config")

    # Step 3: Get Lepton-specific config from deployment.lepton_config
    if not hasattr(cfg.deployment, "lepton_config"):
        raise ValueError(
            "deployment.lepton_config is required when using Lepton executor"
        )

    lepton_config = cfg.deployment.lepton_config

    # Step 4: Convert inference engine config to container spec
    container_spec = _create_inference_container_spec(cfg.deployment)

    # Step 5: Apply Lepton platform deployment configurations
    deployment_config = {
        "resource_requirement": {
            **base_config.get("resource_requirement", {}),
            "resource_shape": lepton_config.resource_shape,
            "min_replicas": lepton_config.min_replicas,
            "max_replicas": lepton_config.max_replicas,
        },
        "auto_scaler": lepton_config.auto_scaler,
        "container": container_spec,
        "envs": [],
    }

    # Add health check configuration if provided
    if hasattr(lepton_config, "health") and lepton_config.health:
        deployment_config["health"] = lepton_config.health
    # Merge deployment config into base config
    final_config = deep_merge(base_config, deployment_config)

    # Step 6: Add environment variables from lepton_config
    if hasattr(lepton_config, "envs") and lepton_config.envs:
        from omegaconf import DictConfig

        for key, value in lepton_config.envs.items():
            env_var: Dict[str, Any] = {"name": key}

            # Support both direct values and secret references
            if isinstance(value, (dict, DictConfig)) and "value_from" in value:
                # Secret reference: {value_from: {secret_name_ref: "secret_name"}}
                env_var["value_from"] = dict(value["value_from"])
            else:
                # Direct value: "direct_value"
                env_var["value"] = str(value)

            final_config["envs"].append(env_var)

    # Step 6b: Auto-populate environment variables from deployment parameters
    _add_deployment_derived_envs(final_config["envs"], cfg.deployment)

    # Step 7: Add mounts with intelligent path construction
    if hasattr(lepton_config, "mounts") and lepton_config.mounts.enabled:
        # Get storage source from task config mounts (since mounts are shared between tasks and deployments)
        storage_source = "node-nfs:lepton-shared-fs"  # default
        if hasattr(cfg, "execution") and hasattr(cfg.execution, "lepton_platform"):
            task_config = cfg.execution.lepton_platform.get("tasks", {})
            task_mounts = task_config.get("mounts", [])
            if task_mounts:
                storage_source = task_mounts[0].get("from", storage_source)

        final_config["mounts"] = [
            {
                "path": lepton_config.mounts.cache_path,
                "from": storage_source,
                "mount_path": lepton_config.mounts.mount_path,
                "mount_options": {},
            }
        ]

    # Step 8: Extract image_pull_secrets to top level (required by Lepton API)
    if "image_pull_secrets" in final_config:
        image_pull_secrets = final_config["image_pull_secrets"]
        # Convert OmegaConf ListConfig to regular Python list
        from omegaconf import ListConfig

        if isinstance(image_pull_secrets, (list, ListConfig)):
            final_config["image_pull_secrets"] = list(image_pull_secrets)
        else:
            # Remove invalid image_pull_secrets
            final_config.pop("image_pull_secrets", None)

    # Step 9: Add API tokens if provided (supports both single and multiple tokens)
    if hasattr(lepton_config, "api_tokens") and lepton_config.api_tokens:
        from omegaconf import DictConfig

        api_tokens_list = []

        for token_config in lepton_config.api_tokens:
            token_var: Dict[str, Any] = {}

            # Support both direct values and secret references
            if isinstance(token_config, (dict, DictConfig)):
                if "value" in token_config:
                    # Direct value: {value: "token_string"}
                    token_var["value"] = str(token_config["value"])
                elif "value_from" in token_config:
                    # Secret reference: {value_from: {secret_name_ref: "secret_name"}}
                    token_var["value_from"] = dict(token_config["value_from"])
            else:
                # Simple string value
                token_var["value"] = str(token_config)

            api_tokens_list.append(token_var)

        final_config["api_tokens"] = api_tokens_list

    # Backward compatibility: support legacy single api_token
    elif hasattr(lepton_config, "api_token") and lepton_config.api_token:
        final_config["api_tokens"] = [{"value": lepton_config.api_token}]

    # Step 10: Replace placeholders
    replacements = {
        "MODEL_CACHE_NAME": _generate_model_cache_name(cfg.deployment.image)
    }
    final_config_with_replacements: Dict[str, Any] = replace_placeholders(
        final_config, replacements
    )

    return final_config_with_replacements


def _create_inference_container_spec(deployment_cfg: DictConfig) -> Dict[str, Any]:
    """Create container specification from inference engine config (vLLM/SGLang/NIM).

    Args:
        deployment_cfg: Deployment configuration containing vLLM/SGLang/NIM settings.

    Returns:
        Container specification for Lepton.
    """
    container_spec = {
        "image": deployment_cfg.image,
        "ports": [{"container_port": deployment_cfg.port}],
    }

    # Generate command based on deployment type
    if deployment_cfg.type == "vllm":
        # Convert vLLM command template to actual command
        command_parts = [
            "vllm",
            "serve",
            deployment_cfg.checkpoint_path,
            f"--tensor-parallel-size={deployment_cfg.tensor_parallel_size}",
            f"--pipeline-parallel-size={deployment_cfg.pipeline_parallel_size}",
            f"--data-parallel-size={deployment_cfg.data_parallel_size}",
            f"--port={deployment_cfg.port}",
            f"--served-model-name={deployment_cfg.served_model_name}",
        ]

        # Add extra args if provided
        if hasattr(deployment_cfg, "extra_args") and deployment_cfg.extra_args:
            command_parts.extend(deployment_cfg.extra_args.split())

        container_spec["command"] = command_parts

    elif deployment_cfg.type == "sglang":
        # Convert SGLang command template to actual command
        command_parts = [
            "python3",
            "-m",
            "sglang.launch_server",
            f"--model-path={deployment_cfg.checkpoint_path}",
            "--host=0.0.0.0",
            f"--port={deployment_cfg.port}",
            f"--served-model-name={deployment_cfg.served_model_name}",
            f"--tp={deployment_cfg.tensor_parallel_size}",
            f"--dp={deployment_cfg.data_parallel_size}",
        ]

        # Add extra args if provided
        if hasattr(deployment_cfg, "extra_args") and deployment_cfg.extra_args:
            command_parts.extend(deployment_cfg.extra_args.split())

        container_spec["command"] = command_parts

    elif deployment_cfg.type == "nim":
        # NIM containers use their default entrypoint - no custom command needed
        # Configuration is handled via environment variables
        pass

    return container_spec


def _add_deployment_derived_envs(envs_list: list, deployment_cfg: DictConfig) -> None:
    """Add environment variables derived from deployment configuration.

    Args:
        envs_list: List to append environment variables to.
        deployment_cfg: Deployment configuration to derive from.
    """
    deployment_type = deployment_cfg.type

    # Common environment variables for all deployment types
    if (
        hasattr(deployment_cfg, "served_model_name")
        and deployment_cfg.served_model_name
    ):
        envs_list.append(
            {"name": "SERVED_MODEL_NAME", "value": deployment_cfg.served_model_name}
        )

    if hasattr(deployment_cfg, "port") and deployment_cfg.port:
        envs_list.append({"name": "MODEL_PORT", "value": str(deployment_cfg.port)})

    # Deployment-specific environment variables
    if deployment_type == "vllm":
        if (
            hasattr(deployment_cfg, "checkpoint_path")
            and deployment_cfg.checkpoint_path
        ):
            envs_list.append(
                {"name": "MODEL_PATH", "value": deployment_cfg.checkpoint_path}
            )
        if (
            hasattr(deployment_cfg, "tensor_parallel_size")
            and deployment_cfg.tensor_parallel_size
        ):
            envs_list.append(
                {
                    "name": "TENSOR_PARALLEL_SIZE",
                    "value": str(deployment_cfg.tensor_parallel_size),
                }
            )

    elif deployment_type == "sglang":
        if (
            hasattr(deployment_cfg, "checkpoint_path")
            and deployment_cfg.checkpoint_path
        ):
            envs_list.append(
                {"name": "MODEL_PATH", "value": deployment_cfg.checkpoint_path}
            )
        if (
            hasattr(deployment_cfg, "tensor_parallel_size")
            and deployment_cfg.tensor_parallel_size
        ):
            envs_list.append(
                {
                    "name": "TENSOR_PARALLEL_SIZE",
                    "value": str(deployment_cfg.tensor_parallel_size),
                }
            )

    elif deployment_type == "nim":
        # NIM-specific derived environment variables
        if (
            hasattr(deployment_cfg, "served_model_name")
            and deployment_cfg.served_model_name
        ):
            envs_list.append(
                {"name": "NIM_MODEL_NAME", "value": deployment_cfg.served_model_name}
            )


def _generate_model_cache_name(image: str) -> str:
    """Generate a cache directory name from the container image.

    Args:
        image: Container image string like 'nvcr.io/nim/meta/llama-3.1-8b-instruct:1.8.6'

    Returns:
        Clean cache name like 'llama-3-1-8b-instruct'
    """
    # Extract model name from image path
    if "/" in image:
        model_part = image.split("/")[-1]  # Get 'llama-3.1-8b-instruct:1.8.6'
    else:
        model_part = image

    # Remove version tag
    if ":" in model_part:
        model_part = model_part.split(":")[0]  # Get 'llama-3.1-8b-instruct'

    # Replace dots with dashes for filesystem compatibility
    return model_part.replace(".", "-")


def create_lepton_endpoint(cfg: DictConfig, endpoint_name: str) -> bool:
    """Create a Lepton endpoint using the lep CLI.

    Args:
        cfg: The nemo-evaluator-launcher configuration object.
        endpoint_name: Name for the endpoint.

    Returns:
        True if endpoint creation succeeded, False otherwise.
    """
    spec = generate_lepton_spec(cfg)

    # Convert OmegaConf objects to regular Python objects for JSON serialization
    from omegaconf import DictConfig, ListConfig

    def convert_to_json_serializable(obj: Any) -> Any:
        """Recursively convert OmegaConf objects to regular Python objects."""
        if isinstance(obj, (DictConfig, dict)):
            return {k: convert_to_json_serializable(v) for k, v in obj.items()}
        elif isinstance(obj, (ListConfig, list)):
            return [convert_to_json_serializable(item) for item in obj]
        else:
            return obj

    json_spec = convert_to_json_serializable(spec)

    # Write spec to temporary file
    import tempfile

    with tempfile.NamedTemporaryFile(mode="w", suffix=".json", delete=False) as f:
        json.dump(json_spec, f, indent=2)
        spec_file = f.name

    try:
        # Create endpoint using lep CLI
        result = subprocess.run(
            ["lep", "endpoint", "create", "--file", spec_file, "--name", endpoint_name],
            capture_output=True,
            text=True,
            timeout=300,
        )

        if result.returncode == 0:
            print(f"✅ Successfully created Lepton endpoint: {endpoint_name}")
            return True
        else:
            error_msg = result.stderr.strip() if result.stderr else ""
            output_msg = result.stdout.strip() if result.stdout else ""
            print(
                f"✗ Failed to create Lepton endpoint | Endpoint: {endpoint_name} | Return code: {result.returncode}"
            )
            if error_msg:
                print(f"   stderr: {error_msg}")
            if output_msg:
                print(f"   stdout: {output_msg}")
            return False

    except subprocess.TimeoutExpired as e:
        print(
            f"✗ Timeout creating Lepton endpoint | Endpoint: {endpoint_name} | Timeout: 300s"
        )
        if hasattr(e, "stderr") and e.stderr:
            print(f"   stderr: {e.stderr}")
        if hasattr(e, "stdout") and e.stdout:
            print(f"   stdout: {e.stdout}")
        return False
    except subprocess.CalledProcessError as e:
        print(
            f"✗ Error creating Lepton endpoint | Endpoint: {endpoint_name} | Error: {e}"
        )
        if hasattr(e, "stderr") and e.stderr:
            print(f"   stderr: {e.stderr}")
        if hasattr(e, "stdout") and e.stdout:
            print(f"   stdout: {e.stdout}")
        return False
    finally:
        # Clean up temporary file
        Path(spec_file).unlink(missing_ok=True)


def delete_lepton_endpoint(endpoint_name: str) -> bool:
    """Delete a Lepton endpoint.

    Args:
        endpoint_name: Name of the endpoint to delete.

    Returns:
        True if deletion succeeded, False otherwise.
    """
    try:
        result = subprocess.run(
            ["lep", "endpoint", "remove", "--name", endpoint_name],
            capture_output=True,
            text=True,
            timeout=60,
        )

        return result.returncode == 0
    except (subprocess.TimeoutExpired, subprocess.CalledProcessError):
        return False


def get_lepton_endpoint_status(endpoint_name: str) -> Optional[dict[str, Any]]:
    """Get the status of a Lepton endpoint.

    Args:
        endpoint_name: Name of the endpoint.

    Returns:
        Status dict if endpoint exists, None otherwise. See
        https://github.com/leptonai/leptonai/blob/7de93b95357126da1e86fa99f54f9a769d5d2646/leptonai/api/v1/types/deployment.py#L338
        for the definition.
    """
    try:
        # TODO(agronskiy): why not use Python API?
        cmd = ["lep", "endpoint", "get", "--name", endpoint_name]
        result = subprocess.run(
            cmd,
            capture_output=True,
            text=True,
            timeout=30,
        )

        if result.returncode != 0:
            return None

        endpoint_info = json.loads(result.stdout)
        status = endpoint_info.get("status", {})
        if isinstance(status, dict):
            return status
        logger.error(
            "Result of running lep command returne non-dict status",
            cmd=cmd,
            status=status,
        )
        return None

    except (
        subprocess.TimeoutExpired,
        subprocess.CalledProcessError,
        json.JSONDecodeError,
    ):
        return None


def wait_for_lepton_endpoint_ready(endpoint_name: str, timeout: int = 600) -> bool:
    """Wait for a Lepton endpoint to become ready.

    Args:
        endpoint_name: Name of the endpoint.
        timeout: Maximum time to wait in seconds.

    Returns:
        True if endpoint becomes ready, False if timeout.
    """
    start_time = time.time()
    while time.time() - start_time < timeout:
        status = get_lepton_endpoint_status(endpoint_name)

        # `get_lepton_endpoint_status` might return `None` if
        # e.g. there was a network error, see definition.
        if status is not None:
            state = status.get("state", "").lower()
            if state == "ready":
                logger.info(
                    "Lepton endpoint is ready",
                    endpoint_name=endpoint_name,
                )

                return True
            elif state in ["failed", "error"]:
                return False

        logger.debug(
            "Waiting for lepton endpoint",
            endpoint_name=endpoint_name,
            timeout=timeout,
            time_delta=time.time() - start_time,
            curr_status=status,
        )
        time.sleep(10)

    logger.error(
        "Timeout waiting for lepton endpoint",
        endpoint_name=endpoint_name,
        timeout=timeout,
    )
    return False


def get_lepton_endpoint_url(endpoint_name: str) -> Optional[str]:
    """Get the URL of a Lepton endpoint.

    Args:
        endpoint_name: Name of the endpoint.

    Returns:
        Endpoint URL if available, None otherwise.
    """
    try:
        result = subprocess.run(
            ["lep", "endpoint", "get", "--name", endpoint_name],
            capture_output=True,
            text=True,
            timeout=30,
        )

        if result.returncode == 0:
            endpoint_info = json.loads(result.stdout)
            status = endpoint_info.get("status", {})
            endpoint = status.get("endpoint", {})
            external_endpoint = endpoint.get("external_endpoint")
            # Ensure we return a proper string type or None
            if isinstance(external_endpoint, str):
                return external_endpoint
            else:
                return None
        else:
            return None
    except (
        subprocess.TimeoutExpired,
        subprocess.CalledProcessError,
        json.JSONDecodeError,
    ):
        return None



# SPDX-FileCopyrightText: Copyright (c) 2025, NVIDIA CORPORATION. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Export evaluation results to specified target."""

from dataclasses import dataclass
from typing import Any, List, Optional

from simple_parsing import field


@dataclass
class ExportCmd:
    """Export evaluation results."""

    # Short usage examples will show up in -h as the class docstring:
    # Examples:
    #   nemo-evaluator-launcher export 8abcd123 --dest local --format json --out .
    #   nemo-evaluator-launcher export 8abcd123.0 9ef01234 --dest local --format csv --out results/ -fname processed_results.csv
    #   nemo-evaluator-launcher export 8abcd123 --dest jet

    invocation_ids: List[str] = field(
        positional=True,
        help="IDs to export (space-separated). Accepts invocation IDs (xxxxxxxx) and job IDs (xxxxxxxx.n); mixture of both allowed.",
    )
    dest: str = field(
        default="local",
        alias=["--dest"],
        choices=["local", "wandb", "mlflow", "gsheets", "jet"],
        help="Export destination.",
    )
    # overrides for exporter config; use -o similar to run command
    override: List[str] = field(
        default_factory=list,
        action="append",
        nargs="?",
        alias=["-o", "--override"],
        help="Hydra-style overrides for exporter config. Use `export.<dest>.key=value` (e.g., -o export.wandb.entity=org-name).",
    )
    output_dir: Optional[str] = field(
        default=".",
        alias=["--output-dir", "-out"],
        help="Output directory (default: current directory).",
    )
    output_filename: Optional[str] = field(
        default=None,
        alias=["--output-filename", "-fname"],
        help="Summary filename (default: processed_results.json/csv based on --format).",
    )
    format: Optional[str] = field(
        default=None,
        alias=["--format"],
        choices=["json", "csv"],
        help="Summary format for --dest local. Omit to only copy artifacts.",
    )
    copy_logs: bool = field(
        default=False,
        alias=["--copy-logs"],
        help="Include logs when copying locally (default: False).",
    )
    log_metrics: List[str] = field(
        default_factory=list,
        alias=["--log-metrics"],
        help="Filter metrics by name (repeatable). Examples: score, f1, mmlu_score_micro.",
    )
    only_required: Optional[bool] = field(
        default=None,
        alias=["--only-required"],
        help="Copy only required+optional artifacts (default: True). Set to False to copy all available artifacts.",
    )

    def execute(self) -> None:
        """Execute export."""
        # Import heavy dependencies only when needed
        from omegaconf import OmegaConf

        from nemo_evaluator_launcher.api.functional import export_results

        # Validation: ensure IDs are provided
        if not self.invocation_ids:
            print("Error: No IDs provided. Specify one or more invocation or job IDs.")
            print(
                "Usage: nemo-evaluator-launcher export <id> [<id>...] --dest <destination>"
            )
            return

        config: dict[str, Any] = {
            "copy_logs": self.copy_logs,
        }

        # Output handling
        if self.output_dir:
            config["output_dir"] = self.output_dir
        if self.output_filename:
            config["output_filename"] = self.output_filename

        # Format and filters
        if self.format:
            config["format"] = self.format
        if self.log_metrics:
            config["log_metrics"] = self.log_metrics

        # Add only_required if explicitly passed via CLI
        if self.only_required is not None:
            config["only_required"] = self.only_required

        # Parse and validate overrides
        if self.override:
            # Flatten possible list-of-lists from parser
            flat_overrides: list[str] = []
            for item in self.override:
                if isinstance(item, list):
                    flat_overrides.extend(str(x) for x in item)
                else:
                    flat_overrides.append(str(item))

            try:
                self._validate_overrides(flat_overrides, self.dest)
            except ValueError as e:
                print(f"Error: {e}")
                return

            # Expand env vars in override vals ($VAR / ${VAR})
            import os

            from omegaconf import OmegaConf

            expanded_overrides: list[str] = []
            for ov in flat_overrides:
                if "=" in ov:
                    k, v = ov.split("=", 1)
                    expanded_overrides.append(f"{k}={os.path.expandvars(v)}")
                else:
                    expanded_overrides.append(os.path.expandvars(ov))

            dot_cfg = OmegaConf.from_dotlist(expanded_overrides)
            as_dict = OmegaConf.to_container(dot_cfg, resolve=True) or {}
            if isinstance(as_dict, dict) and "export" in as_dict:
                export_map = as_dict.get("export") or {}
                if isinstance(export_map, dict) and self.dest in export_map:
                    config.update(export_map[self.dest] or {})
                else:
                    config.update(as_dict)
            else:
                config.update(as_dict)

        if self.format and self.dest != "local":
            print(
                "Note: --format is only used by --dest local. It will be ignored for other destinations."
            )

        if "only_required" in config and self.only_required is True:
            config.pop("only_required", None)

        print(
            f"Exporting {len(self.invocation_ids)} {'invocations' if len(self.invocation_ids) > 1 else 'invocation'} to {self.dest}..."
        )

        result = export_results(self.invocation_ids, self.dest, config)

        if not result.get("success", False):
            err = result.get("error", "Unknown error")
            print(f"\nExport failed: {err}")
            # Provide actionable guidance for common configuration issues
            if self.dest == "mlflow":
                if "tracking_uri" in str(err).lower():
                    print("\nMLflow requires 'tracking_uri' to be configured.")
                    print(
                        "Set it via: -o export.mlflow.tracking_uri=http://mlflow-server:5000"
                    )
                elif "not installed" in str(err).lower():
                    print("\nMLflow package not installed.")
                    print("Install via: pip install nemo-evaluator-launcher[mlflow]")
            elif self.dest == "wandb":
                if "entity" in str(err).lower() or "project" in str(err).lower():
                    print("\nW&B requires 'entity' and 'project' to be configured.")
                    print(
                        "Set via: -o export.wandb.entity=my-org -o export.wandb.project=my-proj"
                    )
                elif "not installed" in str(err).lower():
                    print("\nW&B package not installed.")
                    print("Install via: pip install nemo-evaluator-launcher[wandb]")
            elif self.dest == "gsheets":
                if "not installed" in str(err).lower():
                    print("\nGoogle Sheets package not installed.")
                    print("Install via: pip install nemo-evaluator-launcher[gsheets]")
            return

        # Success path
        if len(self.invocation_ids) == 1:
            # Single invocation
            invocation_id = self.invocation_ids[0]
            print(f"Export completed for {invocation_id}")

            for job_id, job_result in result["jobs"].items():
                if job_result.get("success"):
                    print(f"  {job_id}: {job_result.get('message', '')}")
                    metadata = job_result.get("metadata", {})
                    if metadata.get("run_url"):
                        print(f"    URL: {metadata['run_url']}")
                    if metadata.get("summary_path"):
                        print(f"    Summary: {metadata['summary_path']}")
                    path_hint = job_result.get("dest") or metadata.get("output_dir")
                    if self.dest == "local" and path_hint:
                        print(f"    Path: {path_hint}")
                else:
                    print(f"  {job_id} failed: {job_result.get('message', '')}")
        else:
            # Multiple invocations
            metadata = result.get("metadata", {})
            print(
                f"Export completed: {metadata.get('successful_invocations', 0)}/{metadata.get('total_invocations', 0)} successful"
            )

            # Show summary path if available
            if metadata.get("summary_path"):
                print(f"Summary: {metadata['summary_path']}")
            # Show per-invocation status
            for invocation_id, inv_result in result["invocations"].items():
                if inv_result.get("success"):
                    job_count = len(inv_result.get("jobs", {}))
                    print(f"  {invocation_id}: {job_count} jobs")
                else:
                    print(
                        f"  {invocation_id}: failed, {inv_result.get('error', 'Unknown error')}"
                    )

    def _validate_overrides(self, overrides: List[str], dest: str) -> None:
        """Validate override list for destination consistency.

        Raises:
            ValueError: If overrides specify wrong destination or have other issues.
        """
        if not overrides:
            return  # nothing to validate

        # Check each override for destination mismatch
        for override_str in overrides:
            if override_str.startswith(
                "export."
            ):  # check if override starts with export.
                # Extract destination from override path
                try:
                    key_part = override_str.split("=")[0]  # Get left side before =
                    parts = key_part.split(".")
                    if len(parts) >= 2:
                        override_dest = parts[1]
                        if override_dest != dest:
                            raise ValueError(
                                f"Override destination mismatch: override specifies 'export.{override_dest}' but --dest is '{dest}'. "
                                f"Either change --dest to '{override_dest}' or use 'export.{dest}' in overrides."
                            )
                except (IndexError, AttributeError):
                    # miconstructed override -> OmegaConf handles this
                    pass
